//implementation of UZ file decompressor and saver

//BYTE temp[255];
#include "udemoprivate.h"
#include "FCodec.h"
#include "FConfigCacheIni.h"

IMPLEMENT_CLASS(UUZHandler);
void UUZHandler::execAppend (FFrame& Stack, RESULT_DECL){ 
	guard (UUZHandler::execAppend);
	P_GET_BYTE(Count);
	P_GET_ARRAY_REF(BYTE,B); //get byte array
	P_FINISH;
	//uses a buffered writer for downloading
	if (!(FArchive*)UzAr){ //create buffer writer	
		GLog->Logf(TEXT("loading buffer writer"));
	//	GLog->Logf(TEXT("first 4 bytes: 0x%x-%x-%x-%x"),B[0],B[1],B[2],B[3]);
		UzAr = (INT) (new FByteWriter());
	}
	//((FByteWriter*)UzAr)->Serialize(B,Count);
	((FByteWriter*)UzAr)->Process(B,Count);
	unguard;
}
IMPLEMENT_FUNCTION (UUZHandler,-1,execAppend);

void MoveOldFile(const TCHAR * OldF, UUZHandler* H){ //move any found "oldfile" into the cache!
	H->BeginLoad(); 	
	ULinkerLoad* Linker = H->GetPackageLinker( NULL, OldF, LOAD_None, NULL, NULL); //no cache allowed!
	if (!Linker){
		H->EndLoad();
		return; //is safe to write :)
	}
	const TCHAR * RealF = (*Linker->Filename);
	FGuid GUID = Linker->Summary.Guid;
	delete Linker;
	H->EndLoad();
	if (GFileManager->Move(*(GSys->CachePath + TEXT("\\") + GUID.String()),RealF,1,1)){
		GLog->Logf(TEXT("Successfully moved %s to %s"),RealF,*GUID.String());
		GConfig->SetString(TEXT("Cache"),GUID.String(),RealF,*(GSys->CachePath + TEXT("\\cache.ini")));
	}
	else
		GWarn->Logf(TEXT("FAILED TO move %s to %s"),RealF,*GUID.String());
}

const FString FindDir(const FString FName, UUZHandler* H){ //find the appropriate path for a file!
	FString ext = FName.Mid(FName.InStr(TEXT("."),1));
	for (int i=0;i<GSys->Paths.Num();i++){
		if (GSys->Paths(i).Mid(GSys->Paths(i).InStr(TEXT("*"),1)+1) == ext)
			return (GSys->Paths(i).Left(GSys->Paths(i).InStr(TEXT("*")))); //return path
	}
	return FString();
}

void UUZHandler::execSaveFile (FFrame& Stack, RESULT_DECL){ 
	guard (UUZHandler::execSaveFile);
	P_GET_UBOOL(bCache);
	P_FINISH;
	FBufferReader * Reader;
	guard(ReaderOpen);
	Reader = new FBufferReader(((FByteWriter*)UzAr)->Bytes);
	unguard;
	//BEGIN DECOMPRESSION STAGE:
	INT Signature;
	UBOOL NewVer =false;
    guard(UZSignatureTest);
	*Reader << Signature;   
	if (Signature == 5678)
		NewVer = true;
	else if (Signature != 1234){ //error
		*(BYTE*)Result=3; 
		GWarn->Logf(TEXT("Downloaded file has invalid signature: %i"),Signature);
		delete (FByteWriter*)UzAr; 
		delete Reader; 
		return;
	}
	unguard;
/*	TCHAR * SrcName = TEXT("");
	//have to read names the hard way
	BYTE tsize;
	*Reader << tsize;
	Reader->Serialize(SrcName,tsize);
	SrcName=ANSI_TO_TCHAR((const char*)SrcName);
	FString FileName = FString(SrcName);*/
	*Reader << Filename;
	GLog->Logf(TEXT("Beginning decompression sequence on '%s.uz'"),Filename);
	guard(Decompress);
	FCodecFull Codec; 
    Codec.AddCodec(new FCodecRLE); 
    Codec.AddCodec(new FCodecBWT); 
    Codec.AddCodec(new FCodecMTF); 
    if( NewVer ) 
       Codec.AddCodec(new FCodecRLE); 
    Codec.AddCodec(new FCodecHuffman); 
    UzDeCompAr = (INT) (new FByteWriter()); //decompress archive
	Codec.Decode( *Reader, *(FArchive*)UzDeCompAr ); 
	unguard;
    delete (FByteWriter*)UzAr; 
    delete Reader; 
	guard(PackageTests);
	//check if int file!
	if (Filename.Mid(Filename.InStr(TEXT("."),true)).Locs() == TEXT(".int")){ //is int file.. save to system
		FArchive * OutFile = GFileManager->CreateFileWriter(*Filename,0,GLog);
		OutFile->Serialize(((FByteWriter*)UzDeCompAr)->Bytes.GetData(), ((FArchive*)UzDeCompAr)->TotalSize());
		OutFile->Flush();
		delete OutFile;
		delete (FArchive*)UzDeCompAr; //no longer needed
		*(BYTE*)Result = 0;
		return;
	}
	//NOT INT FILE: BEGIN PACKAGE CHECKING STAGE:
	Reader = new FBufferReader(((FByteWriter*)UzDeCompAr)->Bytes);
	INT		Tag;
	*Reader << Tag;
	if (Tag != 0x9E2A83C1){ //check valid tag
		*(BYTE*)Result=3;
		GWarn->Logf(TEXT("Download file is NOT a valid UTPackage (has tag: %x)"),Tag);
		delete (FArchive*)UzDeCompAr;
		delete Reader;
		return;
	}
	Reader->Seek(36);
	FGuid dGuid;
	*Reader << dGuid;
	if (dGuid != FileGUID){ //bad guid
		*(BYTE*)Result=2;
		delete (FArchive*)UzDeCompAr;
		delete Reader;
		GWarn->Logf(TEXT("Download file's GUID (%s) does not match goal GUID (%s)"),dGuid.String,FileGUID.String());
		return;
	}
	INT GenCount;
	*Reader << GenCount;
	if (GenCount < FileGen){ //bad generation (too old)
		*(BYTE*)Result=1;
		GWarn->Logf(TEXT("Downloaded file has outdated generation (Goal: %i, downloaded: %i)"),FileGen,GenCount);
		delete Reader;
		return; //don't delete: user can "force save"
	}
	unguard;
	((FArchive*)UzDeCompAr)->Seek(0);
	//Begin caching save check
	guard(Saving);
	FArchive * OutFile;
	if (bCache){
		MoveOldFile(*Filename,this);
		OutFile = GFileManager->CreateFileWriter(*(GSys->CachePath + TEXT("\\") + FileGUID.String() + TEXT(".uxx")),0,GLog);
		GLog->Logf(TEXT("Trying to save decompressed file to: %s"),*(GSys->CachePath + TEXT("\\") + FileGUID.String()+ TEXT(".uxx")));
		GConfig->SetString(TEXT("Cache"),FileGUID.String(),*Filename,*(GSys->CachePath + TEXT("\\cache.ini")));
		GConfig->Flush(false, *(GSys->CachePath + TEXT("\\cache.ini")));
	}
	else{
		FString dir = (FindDir(Filename,this));
		if (dir==TEXT("")){ //no dir found!
			delete (FArchive*)UzDeCompAr;
			*(BYTE*)Result=3;
			return;				
		}
		dir += *Filename;
		OutFile = GFileManager->CreateFileWriter(*(dir),0,GLog);
		GLog->Logf(TEXT("Trying to save decompressed file to: %s"),dir );
	}
	if (OutFile==NULL){ //error
		delete (FArchive*)UzDeCompAr;
		GWarn->Logf(TEXT("UNABLE TO SAVE FILE!"));
		*(BYTE*)Result=3;
		return;
	}
	OutFile->Serialize(((FByteWriter*)UzDeCompAr)->Bytes.GetData(), ((FArchive*)UzDeCompAr)->TotalSize());
	OutFile->Flush();
	delete OutFile;
	delete (FArchive*)UzDeCompAr; //no longer needed
	unguard;
	*(BYTE*)Result = 0;
	unguard;
}
IMPLEMENT_FUNCTION (UUZHandler,-1,execSaveFile);

void UUZHandler::execForceSave (FFrame& Stack, RESULT_DECL){ 
	guard (UUZHandler::execForceSave);
	P_GET_UBOOL(bCache);
	P_FINISH;
	FArchive * OutFile;
	if (bCache){
		MoveOldFile(*Filename,this);
		OutFile = GFileManager->CreateFileWriter(*(GSys->CachePath + TEXT("\\") + FileGUID.String() + TEXT(".uxx")),0,GLog);
		GLog->Logf(TEXT("Trying to save decompressed file to: %s"),*(GSys->CachePath + TEXT("\\") + FileGUID.String()+ TEXT(".uxx")));
		GConfig->SetString(TEXT("Cache"),FileGUID.String(),*Filename,*(GSys->CachePath + TEXT("\\cache.ini")));
		GConfig->Flush(false, *(GSys->CachePath + TEXT("\\cache.ini")));
	}
	else{
		FString dir = (FindDir(Filename,this));
		if (dir==TEXT("")){ //no dir found!
			delete (FArchive*)UzDeCompAr;
			*(BYTE*)Result=3;
			return;				
		}
		dir += *Filename;
		OutFile = GFileManager->CreateFileWriter(*(dir),0,GLog);
		GLog->Logf(TEXT("Trying to save decompressed file to: %s"),dir );
	}
	if (OutFile==NULL){ //error
		delete (FArchive*)UzDeCompAr;
		GWarn->Logf(TEXT("UNABLE TO SAVE FILE!"));
		*(BYTE*)Result=3;
		return;
	}
	OutFile->Serialize(((FByteWriter*)UzDeCompAr)->Bytes.GetData(), ((FArchive*)UzDeCompAr)->TotalSize());
	OutFile->Flush();
	delete OutFile;
	delete (FArchive*)UzDeCompAr; //no longer needed
	*(BYTE*)Result = 0;
	unguard;

}
IMPLEMENT_FUNCTION (UUZHandler,-1,execForceSave);